
//**************************************************************************
//**
//** symbol.c
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include <string.h>
#include <stdlib.h>
#include "common.h"
#include "symbol.h"
#include "misc.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

typedef struct
{
	char *name;
	pcd_t directCommand;
	pcd_t stackCommand;
	int argCount;
	int optMask;
	int outMask;
	int invalidMask;
	boolean hasReturnValue;
} internFuncDef_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static symbolNode_t *Find(char *name, symbolNode_t *root);
static symbolNode_t *Insert(char *name, symbolType_t type,
	symbolNode_t **root);
static void FreeNodes(symbolNode_t *root);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static symbolNode_t *LocalRoot;
static symbolNode_t *GlobalRoot;

static internFuncDef_t InternalFunctions[] =
{
	{ "tagwait", PCD_TAGWAITDIRECT, PCD_TAGWAIT, 1, 0, 0, RENDER_MASK, NO },
	{ "polywait", PCD_POLYWAITDIRECT, PCD_POLYWAIT, 1, 0, 0, RENDER_MASK, NO },
	{ "scriptwait", PCD_SCRIPTWAITDIRECT, PCD_SCRIPTWAIT, 1, 0, 0, RENDER_MASK, NO },
	{ "delay", PCD_DELAYDIRECT, PCD_DELAY, 1, 0, 0, RENDER_MASK, NO },
	{ "random", PCD_RANDOMDIRECT, PCD_RANDOM, 2, 0, 0, 0, YES },
	{ "thingcount", PCD_THINGCOUNTDIRECT, PCD_THINGCOUNT, 2, 0, 0, 0, YES },
	{ "changefloor", PCD_CHANGEFLOORDIRECT, PCD_CHANGEFLOOR, 2, 0, 0, RENDER_MASK, NO },
	{ "changeceiling", PCD_CHANGECEILINGDIRECT, PCD_CHANGECEILING, 2, 0, 0, RENDER_MASK, NO },
	{ "lineside", PCD_NOP, PCD_LINESIDE, 0, 0, 0, RENDER_MASK, YES },
	{ "clearlinespecial", PCD_NOP, PCD_CLEARLINESPECIAL, 0, 0, 0, RENDER_MASK, NO },
	{ "playercount", PCD_NOP, PCD_PLAYERCOUNT, 0, 0, 0, 0, YES },
	{ "gametype", PCD_NOP, PCD_GAMETYPE, 0, 0, 0, 0, YES },
	{ "gameskill", PCD_NOP, PCD_GAMESKILL, 0, 0, 0, 0, YES },
	{ "timer", PCD_NOP, PCD_TIMER, 0, 0, 0, 0, YES },
	{ "sectorsound", PCD_NOP, PCD_SECTORSOUND, 2, 0, 0, RENDER_MASK, NO },
	{ "ambientsound", PCD_NOP, PCD_AMBIENTSOUND, 2, 0, 0, RENDER_MASK, NO },
	{ "soundsequence", PCD_NOP, PCD_SOUNDSEQUENCE, 1, 0, 0, RENDER_MASK, NO },
	{ "setlinetexture", PCD_NOP, PCD_SETLINETEXTURE, 4, 0, 0, RENDER_MASK, NO },
	{ "setlineblocking", PCD_NOP, PCD_SETLINEBLOCKING, 2, 0, 0, RENDER_MASK, NO },
	{ "setlinespecial", PCD_NOP, PCD_SETLINESPECIAL, 7, 4|8|16|32|64, 0, RENDER_MASK, NO },
	{ "thingsound", PCD_NOP, PCD_THINGSOUND, 3, 0, 0, RENDER_MASK, NO },
	{ "activatorsound", PCD_NOP, PCD_ACTIVATORSOUND, 2, 0, 0, RENDER_MASK, NO },
	{ "localambientsound", PCD_NOP, PCD_LOCALAMBIENTSOUND, 2, 0, 0, 0, NO },
	{ "setlinemonsterblocking", PCD_NOP, PCD_SETLINEMONSTERBLOCKING, 2, 0, 0, RENDER_MASK, NO },
	{ "fixedmul", PCD_NOP, PCD_FIXEDMUL, 2, 0, 0, 0, YES },
	{ "fixeddiv", PCD_NOP, PCD_FIXEDDIV, 2, 0, 0, 0, YES },
	{ "playerblueskull", PCD_NOP, PCD_PLAYERBLUESKULL, 0, 0, 0, 0, YES },
// [BC] Start of new pcodes
	{ "playerredskull", PCD_NOP, PCD_PLAYERREDSKULL, 0, 0, 0, 0, YES },
	{ "playeryellowskull", PCD_NOP, PCD_PLAYERYELLOWSKULL, 0, 0, 0, 0, YES },
	{ "playermasterskull", PCD_NOP, PCD_PLAYERMASTERSKULL, 0, 0, 0, 0, YES },
	{ "playerbluecard", PCD_NOP, PCD_PLAYERBLUECARD, 0, 0, 0, 0, YES },
	{ "playerredcard", PCD_NOP, PCD_PLAYERREDCARD, 0, 0, 0, 0, YES },
	{ "playeryellowcard", PCD_NOP, PCD_PLAYERYELLOWCARD, 0, 0, 0, 0, YES },
	{ "playermastercard", PCD_NOP, PCD_PLAYERMASTERCARD, 0, 0, 0, 0, YES },
	{ "playerblackskull", PCD_NOP, PCD_PLAYERBLACKSKULL, 0, 0, 0, 0, YES },
	{ "playersilverskull", PCD_NOP, PCD_PLAYERSILVERSKULL, 0, 0, 0, 0, YES },
	{ "playergoldskull", PCD_NOP, PCD_PLAYERGOLDSKULL, 0, 0, 0, 0, YES },
	{ "playerblackcard", PCD_NOP, PCD_PLAYERBLACKCARD, 0, 0, 0, 0, YES },
	{ "playersilvercard", PCD_NOP, PCD_PLAYERSILVERCARD, 0, 0, 0, 0, YES },
	{ "playergoldcard", PCD_NOP, PCD_PLAYERGOLDCARD, 0, 0, 0, 0, YES },
	{ "playerteam1", PCD_NOP, PCD_PLAYERTEAM1, 0, 0, 0, 0, YES },
	{ "playerfrags", PCD_NOP, PCD_PLAYERFRAGS, 0, 0, 0, 0, YES },
	{ "playerhealth", PCD_NOP, PCD_PLAYERHEALTH, 0, 0, 0, 0, YES },
	{ "playerarmorpoints", PCD_NOP, PCD_PLAYERARMORPOINTS, 0, 0, 0, 0, YES },
	{ "playerexpert", PCD_NOP, PCD_PLAYEREXPERT, 0, 0, 0, 0, YES },
	{ "team1count", PCD_NOP, PCD_TEAM1COUNT, 0, 0, 0, 0, YES },
	{ "team2count", PCD_NOP, PCD_TEAM2COUNT, 0, 0, 0, 0, YES },
	{ "team1score", PCD_NOP, PCD_TEAM1SCORE, 0, 0, 0, 0, YES },
	{ "team2score", PCD_NOP, PCD_TEAM2SCORE, 0, 0, 0, 0, YES },
	{ "team1fragpoints", PCD_NOP, PCD_TEAM1FRAGPOINTS, 0, 0, 0, 0, YES },
	{ "music_change", PCD_NOP, PCD_MUSICCHANGE, 2, 0, 0, RENDER_MASK, NO },
// [RH] end of Skull Tag functions
	{ "setgravity", PCD_SETGRAVITYDIRECT, PCD_SETGRAVITY, 1, 0, 0, RENDER_MASK, NO },
	{ "setaircontrol", PCD_SETAIRCONTROLDIRECT, PCD_SETAIRCONTROL, 1, 0, 0, RENDER_MASK, NO },
	{ "clearinventory", PCD_NOP, PCD_CLEARINVENTORY, 0, 0, 0, RENDER_MASK, NO },
	{ "giveinventory", PCD_GIVEINVENTORYDIRECT, PCD_GIVEINVENTORY, 2, 0, 0, RENDER_MASK, NO },
	{ "takeinventory", PCD_TAKEINVENTORYDIRECT, PCD_TAKEINVENTORY, 2, 0, 0, RENDER_MASK, NO },
	{ "checkinventory", PCD_CHECKINVENTORYDIRECT, PCD_CHECKINVENTORY, 1, 0, 0, 0, YES },
	{ "spawn", PCD_SPAWNDIRECT, PCD_SPAWN, 6, 16|32, 0, RENDER_MASK, YES },
	{ "spawnspot", PCD_SPAWNSPOTDIRECT, PCD_SPAWNSPOT, 4, 8|16, 0, RENDER_MASK, YES },
	{ "setmusic", PCD_SETMUSICDIRECT, PCD_SETMUSIC, 3, 2|4, 0, RENDER_MASK, NO },
	{ "localsetmusic", PCD_LOCALSETMUSICDIRECT, PCD_LOCALSETMUSIC, 3, 2|4, 0, RENDER_MASK, NO },
	{ "setstyle", PCD_SETSTYLEDIRECT, PCD_SETSTYLE, 1, 0, 0, ~RENDER_MASK, NO },
	{ "setfont", PCD_SETFONTDIRECT, PCD_SETFONT, 1, 0, 0, 0, NO },
	{ NULL }
};

static char *SymbolTypeNames[] =
{
	"SY_LABEL",
	"SY_SCRIPTVAR",
	"SY_MAPVAR",
	"SY_WORLDVAR",
	"SY_SPECIAL",
	"SY_CONSTANT",
	"SY_INTERNFUNC"
};

// CODE --------------------------------------------------------------------

//==========================================================================
//
// SY_Init
//
//==========================================================================

void SY_Init(void)
{
	symbolNode_t *sym;
	internFuncDef_t *def;

	LocalRoot = NULL;
	GlobalRoot = NULL;
	for(def = InternalFunctions; def->name != NULL; def++)
	{
		sym = SY_InsertGlobal(def->name, SY_INTERNFUNC);
		sym->info.internFunc.directCommand = def->directCommand;
		sym->info.internFunc.stackCommand = def->stackCommand;
		sym->info.internFunc.argCount = def->argCount;
		sym->info.internFunc.optMask = def->optMask;
		sym->info.internFunc.outMask = def->outMask;
		sym->info.internFunc.invalidMask = def->invalidMask;
		sym->info.internFunc.hasReturnValue = def->hasReturnValue;
	}
}

//==========================================================================
//
// SY_Find
//
//==========================================================================

symbolNode_t *SY_Find(char *name)
{
	symbolNode_t *node;

	if((node = SY_FindGlobal(name)) == NULL)
	{
		return SY_FindLocal(name);
	}
	return node;
}

//==========================================================================
//
// SY_FindGlobal
//
//==========================================================================

symbolNode_t *SY_FindGlobal(char *name)
{
	return Find(name, GlobalRoot);
}

//==========================================================================
//
// SY_Findlocal
//
//==========================================================================

symbolNode_t *SY_FindLocal(char *name)
{
	return Find(name, LocalRoot);
}

//==========================================================================
//
// Find
//
//==========================================================================

static symbolNode_t *Find(char *name, symbolNode_t *root)
{
	int compare;
	symbolNode_t *node;

	node = root;
	while(node != NULL)
	{
		compare = strcmp(name, node->name);
		if(compare == 0)
		{
			return node;
		}
		node = compare < 0 ? node->left : node->right;
	}
	return NULL;
}

//==========================================================================
//
// SY_InsertLocal
//
//==========================================================================

symbolNode_t *SY_InsertLocal(char *name, symbolType_t type)
{
	MS_Message(MSG_DEBUG, "Inserting local identifier: %s (%s)\n",
		name, SymbolTypeNames[type]);
	return Insert(name, type, &LocalRoot);
}

//==========================================================================
//
// SY_InsertGlobal
//
//==========================================================================

symbolNode_t *SY_InsertGlobal(char *name, symbolType_t type)
{
	MS_Message(MSG_DEBUG, "Inserting global identifier: %s (%s)\n",
		name, SymbolTypeNames[type]);
	return Insert(name, type, &GlobalRoot);
}

//==========================================================================
//
// SY_InsertGlobalUnique
//
//==========================================================================

symbolNode_t *SY_InsertGlobalUnique(char *name, symbolType_t type)
{
	if(SY_FindGlobal(name) != NULL)
	{ // Redefined
		ERR_Exit(ERR_REDEFINED_IDENTIFIER, YES,
			"Identifier: %s", name);
	}
	return SY_InsertGlobal(name, type);
}

//==========================================================================
//
// Insert
//
//==========================================================================

static symbolNode_t *Insert(char *name, symbolType_t type,
	symbolNode_t **root)
{
	int compare;
	symbolNode_t *newNode;
	symbolNode_t *node;

	newNode = MS_Alloc(sizeof(symbolNode_t), ERR_NO_SYMBOL_MEM);
	newNode->name = MS_Alloc(strlen(name)+1, ERR_NO_SYMBOL_MEM);
	strcpy(newNode->name, name);
	newNode->left = newNode->right = NULL;
	newNode->type = type;
	while((node = *root) != NULL)
	{
		compare = strcmp(name, node->name);
		root = compare < 0 ? &(node->left) : &(node->right);
	}
	*root = newNode;
	return(newNode);
}

//==========================================================================
//
// SY_FreeLocals
//
//==========================================================================

void SY_FreeLocals(void)
{
	MS_Message(MSG_DEBUG, "Freeing local identifiers\n");
	FreeNodes(LocalRoot);
	LocalRoot = NULL;
}

//==========================================================================
//
// SY_FreeGlobals
//
//==========================================================================

void SY_FreeGlobals(void)
{
	MS_Message(MSG_DEBUG, "Freeing global identifiers\n");
	FreeNodes(GlobalRoot);
	GlobalRoot = NULL;
}

//==========================================================================
//
// FreeNodes
//
//==========================================================================

static void FreeNodes(symbolNode_t *root)
{
	if(root == NULL)
	{
		return;
	}
	FreeNodes(root->left);
	FreeNodes(root->right);
	free(root->name);
	free(root);
}
